from __future__ import print_function
import functools
import os
import sys
import yaml
import signal
import subprocess
import errno
from datetime import datetime

from contextlib import contextmanager

SKIP = 100
PASS = 0
FAIL = 1


class TimeoutError(Exception):
    def __init__(self, value="Timed Out"):
        self.value = value


class UnsupportedError(Exception):
    pass


def _as_text(bytestring):
    """Naive conversion of subprocess output to Python string"""
    return bytestring.decode("utf-8", "replace")


def setup_bzr(charm_dir):
    try:
        run_bzr(['whoami'], charm_dir)
    except IOError:
        run_bzr(['whoami', 'amulet@dummy-user.tld'], charm_dir)

    run_bzr(["init"], charm_dir)
    # Set the maximum file size to 0 to avoid skipping any files in the
    # charm.
    run_bzr(['config', 'add.maximum_file_size=0'], charm_dir)


def run_bzr(args, working_dir, env=None):
    """Run a Bazaar command in a subprocess"""
    try:
        p = subprocess.Popen(["bzr"] + args, cwd=working_dir, env=env,
                             stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    except OSError as e:
        if e.errno != errno.ENOENT:
            raise
        raise OSError("bzr not found, do you have Bazaar installed?")
    out, err = p.communicate()
    if p.returncode:
        raise IOError("bzr command failed {!r}:\n"
                      "{}".format(args, _as_text(err)))
    return _as_text(out)


def juju(args, env=None):
    try:
        p = subprocess.Popen(['juju'] + args, env=env, stdout=subprocess.PIPE,
                             stderr=subprocess.PIPE)
    except OSError as e:
        if e.errno != errno.ENOENT:
            raise
        raise OSError("juju not found, do you have Juju installed?")
    out, err = p.communicate()
    if p.returncode:
        raise IOError("juju command failed {!r}:\n"
                      "{}".format(args, _as_text(err)))
    return _as_text(out) if out else None


def timeout_gen(seconds):
    """
    Return a counting generator that raises a :class:`TimeoutError` after
    a number of seconds.

    Note, this is non-preemptive; that is, it will only check for timeout
    between iterations.  This means that it is guaranteed to not timeout
    in the middle of doing its work / checking, which makes it more
    deterministic and easier to debug, but also means that you must ensure
    that the block does not contain an infinite loop or blocking system
    call that needs to be preempted.  If you need a preemptive timeout, see
    :func:`timeout`.

    :param float seconds: Number of seconds after which to timeout.

    Examples::

        for i in timeout(30):
            if i >= 40:
                break  # will timeout first
            sleep(10)

        # wrong!
        for i in timeout(30):
            sleep(60)  # will not preempt! this will take 60s
    """
    start = datetime.now()
    i = 0
    while True:
        yield i
        if (datetime.now() - start).total_seconds() > seconds:
            sys.stderr.write('Timeout occurred, printing juju status...')
            sys.stderr.write(juju(['status']))
            raise TimeoutError()
        i += 1


@contextmanager
def timeout(seconds):
    def signal_handler(signum, frame):
        sys.stderr.write('Timeout occurred, printing juju status...')
        sys.stderr.write(juju(['status']))
        raise TimeoutError()
    signal.signal(signal.SIGALRM, signal_handler)
    signal.alarm(seconds)
    try:
        yield
    finally:
        signal.alarm(0)


class JujuVersion(object):
    def __init__(self, major=0, minor=0, patch=0, get_version=True):
        self.mapping = ['major', 'minor', 'patch']
        self.major = major
        self.minor = minor
        self.patch = patch

        if get_version:
            self.get_version()

    def parse_version(self, version_str):
        version = version_str.split()
        if len(version) > 1:
            version_str = str(version[1])
        else:
            version_str = str(version[0])

        return version_str.split('-')[0].split('.')

    def update_version(self, version_list):
        for i, ver in enumerate(version_list):
            try:
                setattr(self, self.mapping[i], int(ver))
            except:
                break  # List out of range? Versions not semantic? Too bad
        while i < (len(self.mapping) - 1):
            i += 1
            setattr(self, self.mapping[i], None)

    def get_version(self):
        try:
            version = juju(['version'])
        except:
            version = juju(['--version'])

        self.update_version(self.parse_version(version))

    def __str__(self):
        return '.'.join(str(v) for v in [self.major, self.minor, self.patch]
                        if v is not None)


def environments(juju_home=None):
    juju_home = os.path.expanduser(
        juju_home or os.environ.get('JUJU_HOME') or '~/.juju/')
    env_file = os.path.join(juju_home, 'environments.yaml')
    if not os.path.isfile(env_file):
        raise IOError('%s was not found.' % env_file)

    with open(env_file, 'r') as env_yaml:
        envs = yaml.safe_load(env_yaml.read())

    return envs


def raise_status(code, msg=None):
    if msg:
        print(msg)

    sys.exit(code)


def default_environment(juju_home=None):
    juju_home = os.path.expanduser(
        juju_home or os.environ.get('JUJU_HOME') or '~/.juju/')
    envs = environments(juju_home)

    if 'JUJU_ENV' in os.environ:
        return os.environ['JUJU_ENV']

    if os.path.exists(os.path.join(juju_home, 'current-environment')):
        cur_env = None
        with open(os.path.join(juju_home, 'current-environment')) as f:
            cur_env = f.read().strip()

        if cur_env in envs['environments']:
            return cur_env

    if 'default' in envs:
        return envs['default']
    else:
        if len(envs['environments']) != 1:
            raise ValueError('No default environment specified.')

        return next(iter(envs['environments'].keys()))


class reify(object):
    def __init__(self, func):
        self.func = func
        try:
            functools.update_wrapper(self, func)
        except:
            pass

    def __get__(self, inst, obtype=None):
        if inst is None:
            return self

        out = self.func(inst)
        setattr(inst, self.func.__name__, out)
        return out


@contextmanager
def fail_if_timeout(seconds):
    try:
        yield
    except TimeoutError:
        message = 'Unable to set up environment in %d seconds.' % seconds
        raise_status(FAIL, msg=message)
    except:
        raise



import base64
import contextlib
import json
import logging
import os
import shlex
import subprocess
import yaml

from path import path
from path import tempdir

from .helpers import default_environment, juju, timeout as unit_timesout
from .sentry import Talisman
from .charm import CharmCache

logger = logging.getLogger(__name__)


def get_charm_name(dir_):
    """Given a directory, return the name of the charm in that dir.

    If metadata.yaml exists in the dir, grab the charm name from there.
    Otherwise, return the name of the dir.

    """
    try:
        with open(os.path.join(dir_, 'metadata.yaml')) as f:
            return yaml.load(f)['name']
    except:
        return os.path.basename(dir_)


class Deployment(object):
    """A Juju workload.

    Use this class to add, configure, relate, and deploy services to a
    Juju environment.

    :ivar Deployment.sentry: A :class:`amulet.sentry.Talisman` instance that
        becomes available after a call to :meth:`setup` (before that,
        :attr:`sentry` is ``None``.

    """
    log = logger

    def __init__(self, juju_env=None, series='precise',
                 juju_deployer='juju-deployer', **kw):
        """Initialize a deployment.

        :param juju_env: Name of the Juju enviroment in which to deploy. If
            None, the default environment is used.
        :param series: The default series of charms to deploy.
        :param juju_deployer: Path to juju_deployer binary to use for the
            deployment.

        """
        self.services = {}
        self.relations = []
        self.interfaces = []
        self.subordinates = []
        self.series = series
        self.deployed = False
        self.juju_env = juju_env or default_environment()
        self.charm_name = get_charm_name(os.getcwd())

        self.sentry = None
        self.deployer = path(juju_deployer)

        if 'JUJU_TEST_CHARM' in os.environ:
            self.charm_name = os.environ['JUJU_TEST_CHARM']

        self.charm_cache = CharmCache(self.charm_name)

    @classmethod
    def from_bundle(cls, bundle_file, deployment_name=None):
        """Create a :class:`Deployment` object from a bundle file.

        :param bundle_file: Path to the bundle file.
        :param deployment_name: Name of the deployment to use. Useful for
            old-style bundle files that contain multiple named deployments.
        :return: A new :class:`Deployment` object.

        """
        deployment = cls()
        bundle_file = path(bundle_file)
        deployment.load_bundle_file(bundle_file, deployment_name)
        return deployment

    def load_bundle_file(self, bundle_file, deployment_name=None):
        """Load a bundle file from disk.

        :param bundle_file: Path to the bundle file.
        :param deployment_name: Name of the deployment to use. Useful for
            old-style bundle files that contain multiple named deployments.

        """
        with open(bundle_file, 'r') as stream:
            contents = yaml.safe_load(stream)
        return self.load(contents, deployment_name)

    def load(self, deploy_cfg, deployment_name=None):
        """Load an existing deployment schema (bundle) dictionary.

        :param deploy_cfg: The bundle dictionary.
        :param deployment_name: Name of the deployment to use. Useful for
            old-style bundle files that contain multiple named deployments.

        """
        if deployment_name is None and 'services' in deploy_cfg:
            # v4 format
            schema = deploy_cfg
        else:
            # v3 format
            schema = deploy_cfg.get(deployment_name, None) \
                or next(iter(deploy_cfg.values()))
        self.series = schema.get('series', self.series)
        self.relations = schema.get('relations', [])
        for service, service_config in schema['services'].items():
            constraints = service_config.get('constraints')
            if constraints:
                constraints = dict(
                    constraint.split('=')
                    for constraint in constraints.split()
                )

            self.add(
                service,
                charm=service_config.get('charm'),
                units=service_config.get('num_units', 1),
                branch=service_config.get('branch', None),
                constraints=constraints,
                placement=service_config.get('to', None),
                series=self.series
            )

            if service_config.get('options'):
                self.configure(service, service_config['options'])

            if service_config.get('expose'):
                self.expose(service)

    def add(self, service_name,
            charm=None,
            units=1,
            constraints=None,
            branch=None,
            placement=None,
            series=None):
        """Add a new service to the deployment schema.

        :param service_name: Name of the service to deploy.
        :param charm: Name of the charm to deploy for this service. If None,
            defaults to ``service_name``.
        :param units: Number of units to deploy.
        :param constraints: Dictionary of service constraints.
        :param branch: TODO
        :param placement: Placement directive for this service. Examples:

            "1" - Deploy to machine 1
            "lxc:1" - Deploy to lxc container on machine 1
            "lxc:wordpress/0 - Deploy to lxc container on first wordpress unit

        :param series: Series of charm to deploy, e.g. precise, trusty, xenial

        Example::

            import amulet
            d = amulet.Deployment()
            d.add('wordpress')
            d.add('second-wp', charm='wordpress')
            d.add('personal-wp', charm='~marcoceppi/wordpress', units=2)

        """
        if self.deployed:
            raise NotImplementedError('Environment already setup')

        if service_name in self.services:
            raise ValueError('Service is already set to be deployed')

        service = self.services[service_name] = {}

        charm = self.charm_cache.fetch(
            service_name, charm, branch=branch, series=series or self.series)

        if charm.subordinate:
            for rtype in ['provides', 'requires']:
                try:
                    rels = getattr(charm, rtype)
                    for relation in rels:
                        rel = rels[relation]
                        if 'scope' in rel and rel['scope'] == 'container':
                            self.subordinates.append('%s:%s' %
                                                     (service_name, relation))
                except:  # @@ why is this diaper here?
                    pass

        source = charm.url and {'charm': charm.url} \
            or {'branch': charm.code_source['location']}

        service.update(source)

        service['num_units'] = units
        if placement is not None:
            service['to'] = placement

        if 'JUJU_TEST_CONSTRAINTS' in os.environ:
            env_constraints = {}
            for c in os.environ['JUJU_TEST_CONSTRAINTS'].split():
                try:
                    k, v = c.split('=')
                    env_constraints[k] = v
                except:
                    raise ValueError(
                        'Invalid constraint in JUJU_TEST_CONSTRAINTS: '
                        '%s' % (c))
            if constraints is not None:
                env_constraints.update(constraints)
            constraints = env_constraints

        if constraints is not None:
            if not isinstance(constraints, dict):
                raise ValueError('Constraints must be specified as a dict')

            r = ['%s=%s' % (K, V) for K, V in constraints.items()]
            service['constraints'] = " ".join(r)

        return service

    def add_unit(self, service, units=1, target=None):
        """Add more units of an existing service after deployment.

        :param service: Name of service to which units will be added.
        :param units: Number of units to add.
        :param target: Placement directive for the added unit(s).

        Example::

            import amulet
            d = amulet.Deployment()
            d.add('wordpress')
            try:
                d.setup(timeout=900)
            except amulet.helpers.TimeoutError:
                # Setup didn't complete before timeout
                pass
            d.add_unit('wordpress')
            d.add_unit('wordpresss', units=2)
            d.add_unit('wordpresss', target="lxc:1")

        """
        if not isinstance(units, int) or units < 1:
            raise ValueError('Only positive integers can be used for units')
        if target is not None and units != 1:
            raise ValueError(
                "Can't deploy more than one unit when specifying a target.")
        if service not in self.services:
            raise ValueError('Service needs to be added before you can scale')

        self.services[service]['num_units'] = \
            self.services[service].get('num_units', 1) + units

        if self.deployed:
            args = ['add-unit', service, '-n', str(units)]
            if target is not None:
                args.extend(["--to", target])
            juju(args)
            self.sentry = Talisman(self.services)

    def remove_unit(self, *units):
        """Remove (destroy) one or more already-deployed units.

        :param units: One or more units in the form <service>/<unit_num>,
            e.g. "wordpress/0", passed as \*args.

        """
        if not self.deployed:
            raise NotImplementedError('Environment not setup yet')
        if not units:
            raise ValueError('No units provided')
        for unit in units:
            if '/' not in unit:
                raise ValueError('%s is not a unit' % unit)
            service = unit.split('/')[0]
            if service not in self.services:
                raise ValueError('%s is not a deployed service' % service)

        juju(['remove-unit'] + list(units))

        for unit in units:
            if self.sentry and unit in self.sentry.unit:
                del self.sentry.unit[unit]
            service = unit.split('/')[0]
            self.services[service]['num_units'] = \
                min(0, self.services[service].get('num_units', 1) - 1)
    destroy_unit = remove_unit

    def remove_service(self, *services):
        """Remove (destroy) one or more already-deployed services.

        :param services: One or more service names passed as \*args.

        """
        if not services:
            raise ValueError('No services provided')
        for service in services:
            if service not in self.services:
                raise ValueError('%s is not a deployed service' % service)

        if self.deployed:
            juju(['remove-service'] + list(services))

        for service in services:
            self._remove_service_sentries(service)
            self._remove_service_relations(service)
            del self.services[service]
    destroy_service = remove_service

    def remove(self, *units_or_services):
        """Remove (destroy) one or more already-deployed services or units.

        :param units_or_services: One or more service or unit names passed
            as \*args.

        """
        if not units_or_services:
            raise ValueError('No units or services provided')

        units_or_services = set(units_or_services)
        units = {s for s in units_or_services if '/' in s}
        services = units_or_services - units
        units = {u for u in units if u.split('/')[0] not in services}

        if units:
            self.remove_unit(*units)

        if services:
            self.remove_service(*services)
    destroy = remove

    def _remove_service_sentries(self, service):
        if not self.sentry:
            return

        for unit in list(self.sentry.unit):
            if unit.split('/')[0] == service:
                del self.sentry.unit[unit]

    def _remove_service_relations(self, service):
        for relation in self.relations[:]:
            for rel_service in relation:
                if rel_service.split(':')[0] == service:
                    self.relations.remove(relation)
                    break

    def relate(self, *args):
        """Relate two or more services together.

        If more than two arguments are given, the first service is related
        to each of the others.

        :param args: Services to relate, in the form
            "service_name:relation_name".

        Example::

            import amulet
            d = amulet.Deployment()
            d.add('postgresql')
            d.add('mysql')
            d.add('wordpress')
            d.add('mediawiki')
            d.add('discourse')
            d.relate('postgresql:db-admin', 'discourse:db')
            d.relate('mysql:db', 'wordpress:db', 'mediawiki:database')
            # previous command is equivalent too:
            d.relate('mysql:db', 'wordpress:db')
            d.relate('mysql:db', 'mediawiki:database')

        """
        if len(args) < 2:
            raise LookupError('Need at least two services:relation')

        for srv_rel in args:
            if not ':' in srv_rel:
                raise ValueError('All relations must be explicit, ' +
                                 'service:relation')

            srv, rel = srv_rel.split(':')
            if srv not in self.services:
                raise ValueError('Can not relate, service not deployed yet')

            c = self.charm_cache[srv]
            if rel not in list(c.provides.keys()) + list(c.requires.keys()) \
                    + ['juju-info']:
                raise ValueError('%s does not exist for %s' % (rel, srv))

        args = list(args)
        first = args.pop(0)
        for srv in args:
            self._relate(first, srv)

    def _relate(self, a, b):
        if [a, b] not in self.relations and [b, a] not in self.relations:
            self.relations.append([a, b])
            if self.deployed:
                juju(['add-relation'] + [a, b])

    def unrelate(self, *args):
        """Remove a relation between two services.

        :param args: Services to unrelate, in the form
            "service_name:relation_name".

        Example::

            import amulet
            d = amulet.Deployment()
            d.add('postgresql')
            d.add('mysql')
            d.add('wordpress')
            d.add('mediawiki')
            d.add('discourse')
            d.relate('postgresql:db-admin', 'discourse:db')
            d.relate('mysql:db', 'wordpress:db', 'mediawiki:database')
            # unrelate all the services we just related
            d.unrelate('postgresql:db-admin', 'discourse:db')
            d.unrelate('mysql:db', 'wordpress:db')
            d.unrelate('mysql:db', 'mediawiki:database')

        """
        if len(args) != 2:
            raise LookupError('Need exactly two service:relations')

        for srv_rel in args:
            if not ':' in srv_rel:
                raise ValueError('All relations must be explicit, ' +
                                 'service:relation')
        relation = list(args)
        for rel in relation, reversed(relation):
            if rel in self.relations:
                relation = rel
                break
        else:
            raise ValueError('Relation does not exist')

        self.relations.remove(relation)
        if self.deployed:
            juju(['remove-relation'] + relation)

    def schema(self):
        """Return the deployment schema (bundle) as a dictionary.

        """
        return self._deployer_map(self.services, self.relations)

    def configure(self, service, options):
        """Change configuration options for a service (deployed or not).

        :param service: Name of service to configure.
        :param options: Dictionary of configuration settings.

        Example::

            import amulet
            d = amulet.Deployment()
            d.add('postgresql')
            d.configure('postgresql', {
                'autovacuum': True,
                'cluster_name': 'cname',
            })

        """
        for k, v in options.items():
            include_token = 'include-base64://'
            if type(v) is str and v.startswith(include_token):
                v = v.replace(include_token, '')
                with open(os.path.join(os.getcwd(), 'tests', v)) as f:
                    v = base64.b64encode(f.read())
                service['options'][k] = v

        if self.deployed:
            opts = ['set', service]
            for k, v in options.items():
                opts.append("%s=%s" % (k, v))
            return juju(opts)

        if service not in self.services:
            raise ValueError('Service has not yet been described')

        if not 'options' in self.services[service]:
            self.services[service]['options'] = options
        else:
            self.services[service]['options'].update(options)

    def expose(self, service):
        """Expose a service.

        If the service is already deployed it will be exposed immediately,
        otherwise it will be exposed when deployed.

        :param service: Name of the service to expose.

        Example::

            import amulet
            d = amulet.Deployment()
            d.add('varnish')
            d.expose('varnish')

        """
        if self.deployed:
            return juju(['expose', service])

        if service not in self.services:
            raise ValueError('%s has not yet been described' % service)
        self.services[service]['expose'] = True

    @contextlib.contextmanager
    def _deploy_w_timeout(self, timeout):
        """Sets timeout and tmp working directory for wrapped block.

        If successful, sets instance.deployed.

        :param timeout: Amount of time to wait for deployment to complete.

        """
        deploy_dir = tempdir(prefix='amulet_deployment_')
        with deploy_dir, unit_timesout(timeout):
            yield
        self.deployed = True

    def action_defined(self, service):
        """Return list of actions defined for the service.

        :param service: Name of service for which to list actions.
        :return: List of actions, as json.

        """
        if service not in self.services:
            raise ValueError(
                'Service needs to be added before listing actions.')
        raw = juju(['action', 'defined', service, '--format', 'json'])
        return json.loads(raw)

    def action_do(self, unit, action, action_args={}):
        """Run action on a unit and return the result UUID.

        :param unit: Unit on which to run action, e.g. "wordpress/0"
        :param action: Name of action to run.
        :param action_args: Dictionary of action parameters.

        """
        if '/' not in unit:
            raise ValueError('%s is not a unit' % unit)
        cmd = ['action', 'do', unit, action, '--format', 'json']
        for key, value in action_args.items():
            cmd += ["%s=%s" % (str(key), str(value))]
        result = juju(cmd)
        action_result = json.loads(result)
        results_id = action_result["Action queued with id"]
        return results_id

    def action_fetch(self, action_id, timeout=600):
        """Fetch results for an action.

        If the timeout expires and the action is still not complete, an
        empty dictionary is returned.

        :param action_id: UUID of the action.
        :param timeout: Length of time to wait for an action to complete.
        :return: Action results, as json.

        """
        cmd = ['action', 'fetch', action_id, '--format', 'json']
        if timeout is not None:
            cmd += ["--wait", str(timeout)]
        raw = juju(cmd)
        result = json.loads(raw)
        status = result['status']
        if status == 'completed':
            if 'results' in result:
                return result['results']
        return {}

    def setup(self, timeout=600, cleanup=True):
        """Deploy the workload.

        If timeout expires before the deployment completes, raises
            :class:`amulet.helpers.TimeoutError`.

        :param timeout: Amount of time to wait for deployment to complete.
            If environment variable AMULET_SETUP_TIMEOUT is set, it overrides
            this value.
        :param cleanup: Set to False to leave the generated deployer file
            on disk. Useful for debugging.

        Example::

            import amulet
            d = amulet.Deployment()
            d.add('wordpress')
            d.add('mysql')
            d.configure('wordpress', debug=True)
            d.relate('wordpress:db', 'mysql:db')
            try:
                d.setup(timeout=900)
            except amulet.helpers.TimeoutError:
                # Setup didn't complete before timeout
                pass

        """
        timeout = int(os.environ.get('AMULET_SETUP_TIMEOUT') or timeout)

        if not self.deployer:
            raise NameError('Path to juju-deployer is not defined.')

        with tempdir(prefix='amulet-juju-deployer-') as tmpdir:
            schema_json = json.dumps(self.schema(), indent=2)
            self.log.debug("Deployer schema\n%s", schema_json)

            schema_file = tmpdir / 'deployer-schema.json'
            schema_file.write_text(schema_json)

            cmd = "{deployer} -W {debug} -c {schema} -e {env} -t {timeout} {env}"
            cmd_args = dict(
                deployer=self.deployer.expanduser(),
                debug=(
                    '-d'
                    if self.log.getEffectiveLevel() == logging.DEBUG
                    else ''),
                schema=schema_file,
                env=self.juju_env,
                timeout=str(timeout + 100),
            )
            cmd = cmd.format(**cmd_args)
            self.log.debug(cmd)

            with self._deploy_w_timeout(timeout):
                subprocess.check_call(shlex.split(cmd))

        self.sentry = Talisman(self.services, timeout=timeout)
        if cleanup is False:
            tmpdir.makedirs()
            (tmpdir / 'deployer-schema.json').write_text(schema_json)

    def _deployer_map(self, services, relations):
        return {
            self.juju_env: {
                'series': self.series,
                'services': self.services,
                'relations': self._build_relations()
            }
        }

    def _build_relations(self):
        relations = []
        for rel in self.relations:
            relations.append(rel)

        return relations



import re
import unittest
import yaml
from datetime import datetime
from copy import deepcopy

from amulet.sentry import (
    Talisman,
    UnitSentry,
    StatusMessageMatcher,
)
from amulet.helpers import (
    TimeoutError,
    UnsupportedError,
)
from mock import patch, Mock


mock_status = yaml.load("""\
machines:
  "0":
    agent-state: started
    agent-version: 1.24.6.1
    dns-name: localhost
    instance-id: localhost
    series: trusty
    state-server-member-status: has-vote
  "1":
    agent-state: started
    instance-id: johnsca-local-machine-1
    series: trusty
    hardware: arch=amd64
  "2":
    agent-state: started
    instance-id: johnsca-local-machine-1
    series: trusty
    hardware: arch=amd64
  "3":
    agent-state: pending
    instance-id: johnsca-local-machine-1
    series: trusty
    hardware: arch=amd64
services:
  meteor:
    units:
      meteor/0:
        public-address: 10.0.3.152
        workload-status:
          current: active
          message: ready
          since: 24 Sep 2015 16:44:44-04:00
        agent-status:
          current: idle
          since: 24 Sep 2015 16:44:44-04:00
        machine: "1"
      meteor/1:
        public-address: 10.0.3.177
        workload-status:
          current: active
          message: ready
          since: 24 Sep 2015 16:44:44-04:00
        agent-status:
          current: idle
          since: 24 Sep 2015 16:44:44-04:00
        machine: "2"
        subordinates:
          rsyslog-forwarder/0:
            public-address: 10.0.3.115
            workload-status:
              current: active
              message: rsyslog
  relation-sentry:
    units:
      relation-sentry/0:
        public-address: 10.0.3.92
  rsyslog-forwarder:
    charm: cs:trusty/rsyslog-forwarder
    subordinate-to:
    - meteor
    relations:
      juju-info:
      - meteor
  pending:
    units:
      pending/0:
        public-address: 10.0.3.152
        workload-status:
          current: unknown
          since: 24 Sep 2015 16:44:44-04:00
        agent-status:
          current: allocating
          since: 24 Sep 2015 16:44:44-04:00
        machine: "3"
  nopublic:
    units:
      nopublic/0:
        workload-status:
          current: maintainence
          message: working
          since: 24 Sep 2015 16:44:44-04:00
        agent-status:
          current: executing
          since: 24 Sep 2015 16:44:44-04:00
        machine: "2"
  errord:
    units:
      errord/0:
        public-address: 10.0.3.152
        workload-status:
          current: error
          message: 'hook failed: "install"'
          since: 24 Sep 2015 16:44:44-04:00
        agent-status:
          current: idle
          since: 24 Sep 2015 16:44:44-04:00
        machine: "2"
  old:
    units:
      old/0:
        public-address: 10.0.3.152
        agent-state: started
        machine: "2"
  olderrord:
    units:
      olderrord/0:
        public-address: 10.0.3.152
        agent-state: error
        agent-state-info: 'hook failed: "install"'
        machine: "2"
  ubuntu:
    charm: cs:trusty/ubuntu-4
    exposed: false
    service-status:
      current: unknown
      message: Waiting for agent initialization to finish
      since: 30 Sep 2015 16:44:09-04:00
    relations:
      juju-info:
      - sub
      - subpend
    units:
      ubuntu/0:
        workload-status:
          current: unknown
          since: 24 Sep 2015 16:44:44-04:00
        agent-status:
          current: started
          since: 24 Sep 2015 16:44:44-04:00
        agent-state: pending
        machine: "2"
        subordinates:
          sub/0:
            public-address: 10.0.3.115
            agent-status:
              current: started
              since: 24 Sep 2015 16:44:44-04:00
  sub:
    charm: cs:trusty/sub-1
    subordinate-to:
    - ubuntu
    relations:
      juju-info:
      - ubuntu
  subpend:
    charm: cs:trusty/subpend-1
    subordinate-to:
    - ubuntu
    relations:
      juju-info:
      - ubuntu
  unsub:
    charm: cs:trusty/unsub-1
""")


class TalismanTest(unittest.TestCase):
    timeout = 0.01

    @patch.object(Talisman, 'wait_for_status')
    @patch.object(UnitSentry, 'upload_scripts')
    @patch('amulet.sentry.helpers.default_environment')
    def test_init(self, default_env, upload_scripts, wait_for_status):
        default_env.return_value = 'local'
        wait_for_status.return_value = mock_status

        sentry = Talisman(['meteor'], timeout=self.timeout)

        self.assertTrue('meteor/0' in sentry.unit)
        self.assertTrue('meteor/1' in sentry.unit)

    @patch.object(Talisman, 'wait_for_status')
    @patch.object(UnitSentry, 'upload_scripts')
    @patch('amulet.sentry.helpers.default_environment')
    def test_getitem(self, default_env, upload_scripts, wait_for_status):
        default_env.return_value = 'local'
        wait_for_status.return_value = mock_status

        sentry = Talisman(['meteor'], timeout=self.timeout)

        self.assertEqual(sentry['meteor/0'], sentry.unit['meteor/0'])
        self.assertSetEqual(set(sentry['meteor']), set(sentry.unit.values()))

        with self.assertRaises(KeyError):
            sentry['meteor/99']  # Non-existent unit.

        # Non-existent service, or service with no units.
        self.assertEqual(sentry['invalid'], [])

        with self.assertRaises(KeyError):
            sentry['invalid/99']  # Non-existent unit in non-existent service.

    @patch.object(Talisman, 'wait_for_status')
    @patch.object(UnitSentry, 'upload_scripts')
    @patch('amulet.sentry.helpers.default_environment')
    def test_subordinate(self, default_env, upload_scripts, wait_for_status):
        default_env.return_value = 'local'
        wait_for_status.return_value = mock_status

        sentry = Talisman(['meteor', 'rsyslog-forwarder'], timeout=self.timeout)

        self.assertTrue('rsyslog-forwarder/0' in sentry.unit)

    @patch.object(Talisman, '__init__', Mock(return_value=None))
    @patch('amulet.helpers.juju', Mock(return_value='status'))
    @patch('amulet.waiter.status')
    def test_wait_for_status(self, status):
        status.return_value = mock_status
        talisman = Talisman([], timeout=self.timeout)

        talisman.wait_for_status('env', ['meteor'], self.timeout)
        talisman.wait_for_status('env', ['old'], self.timeout)
        talisman.wait_for_status('env', ['sub'], self.timeout)
        talisman.wait_for_status('env', ['unsub'], self.timeout)

        self.assertRaises(TimeoutError, talisman.wait_for_status, 'env', ['pending'], self.timeout)
        self.assertRaises(TimeoutError, talisman.wait_for_status, 'env', ['nopublic'], self.timeout)
        self.assertRaisesRegexp(Exception, r'Error on unit.*hook failed',
                                talisman.wait_for_status, 'env', ['errord'], self.timeout)
        self.assertRaisesRegexp(Exception, r'Error on unit.*hook failed',
                                talisman.wait_for_status, 'env', ['olderrord'], self.timeout)
        self.assertRaises(TimeoutError, talisman.wait_for_status, 'env', ['subpend'], self.timeout)

    @patch.object(Talisman, '__init__', Mock(return_value=None))
    @patch('amulet.helpers.juju', Mock(return_value='status'))
    @patch('amulet.waiter.status')
    def test_wait_for_status_containers(self, status):
        status.return_value = {
            'machines': {
                '0': {
                    'containers': {
                        '0/lxc/1': {'agent-state': 'started'},
                        '0/lxc/2': {'agent-state': 'pending'},
                    },
                },
            },
            'services': {
                'foo': {
                    'units': {
                        'foo/0': {
                            'machine': '0/lxc/1',
                            'public-address': 'something',
                        },
                    },
                },
                'bar': {
                    'units': {
                        'bar/0': {
                            'machine': '0/lxc/2',
                            'public-address': 'something',
                        },
                    },
                },
            },
        }
        talisman = Talisman([], timeout=self.timeout)
        talisman.wait_for_status('env', ['foo'], self.timeout)
        self.assertRaises(TimeoutError, talisman.wait_for_status, 'env', ['bar'], self.timeout)

    @patch('amulet.helpers.juju', Mock(return_value='status'))
    @patch('amulet.helpers.default_environment', Mock())
    @patch.object(UnitSentry, 'upload_scripts', Mock())
    @patch.object(UnitSentry, 'juju_agent')
    @patch('amulet.waiter.status')
    def test_wait(self, _status, juju_agent):
        status = _status.return_value = deepcopy(mock_status)

        def set_state(which, key, value):
            status['services']['meteor']['units']['meteor/0'][which][key] = value

        t = Talisman(['meteor'], timeout=self.timeout)
        t.wait(self.timeout)

        set_state('workload-status', 'current', 'unknown')
        t.wait(self.timeout)

        set_state('workload-status', 'current', 'blocked')
        t.wait(self.timeout)

        set_state('workload-status', 'current', 'active')
        set_state('agent-status', 'current', 'executing')
        self.assertRaises(TimeoutError, t.wait, self.timeout)

        set_state('agent-status', 'current', 'idle')
        set_state('agent-status', 'since', datetime.now().strftime('%d %b %Y %H:%M:%S'))
        self.assertRaises(TimeoutError, t.wait, self.timeout)

        t = Talisman(['old'], timeout=self.timeout)
        juju_agent.return_value = None
        self.assertRaises(TimeoutError, t.wait, self.timeout)

        juju_agent.return_value = {'hook': 'foo'}
        self.assertRaises(TimeoutError, t.wait, self.timeout)

        juju_agent.return_value = {}
        t.wait(self.timeout)

    @patch('amulet.helpers.juju', Mock(return_value='status'))
    @patch('amulet.helpers.default_environment', Mock())
    @patch('amulet.waiter.status')
    def test_wait_for_messages(self, _status):
        status = _status.return_value = deepcopy(mock_status)
        t = Talisman([], timeout=self.timeout)

        def set_status(unit, message):
            service = unit.split('/')[0]
            status['services'][service]['units'][unit]['workload-status']['message'] = message

        t.wait_for_messages({'meteor': 'ready'}, self.timeout)
        t.wait_for_messages({'meteor': re.compile('r..dy')}, self.timeout)
        t.wait_for_messages({'meteor': {'ready'}}, self.timeout)
        t.wait_for_messages({'meteor': re.compile('ready|ok')}, self.timeout)
        self.assertRaises(TimeoutError, t.wait_for_messages, {'meteor': {'ready', 'ok'}}, self.timeout)
        self.assertRaises(TimeoutError, t.wait_for_messages, {'meteor': ['ready', 'ok']}, self.timeout)

        set_status('meteor/0', 'ok')
        self.assertRaises(TimeoutError, t.wait_for_messages, {'meteor': 'ready'}, self.timeout)
        self.assertRaises(TimeoutError, t.wait_for_messages, {'meteor': re.compile('r..dy')}, self.timeout)
        t.wait_for_messages({'meteor': {'ready'}}, self.timeout)
        t.wait_for_messages({'meteor': re.compile('ready|ok')}, self.timeout)
        t.wait_for_messages({'meteor': {'ready', 'ok'}}, self.timeout)
        t.wait_for_messages({'meteor': ['ready', 'ok']}, self.timeout)

        set_status('meteor/1', 'ok')
        self.assertRaises(TimeoutError, t.wait_for_messages, {'meteor': 'ready'}, self.timeout)
        self.assertRaises(TimeoutError, t.wait_for_messages, {'meteor': re.compile('r..dy')}, self.timeout)
        self.assertRaises(TimeoutError, t.wait_for_messages, {'meteor': {'ready'}}, self.timeout)
        t.wait_for_messages({'meteor': re.compile('ready|ok')}, self.timeout)
        self.assertRaises(TimeoutError, t.wait_for_messages, {'meteor': {'ready', 'ok'}}, self.timeout)
        self.assertRaises(TimeoutError, t.wait_for_messages, {'meteor': ['ready', 'ok']}, self.timeout)

        set_status('meteor/0', 'ready')
        status['services']['meteor']['units']['meteor/2'] = {
            'workload-status': {
                'current': 'active',
                'message': 'working',
            },
        }
        t.wait_for_messages({'meteor': {'ready', 'ok'}})
        self.assertRaises(TimeoutError, t.wait_for_messages, {'meteor': ['ready', 'ok']}, self.timeout)
        self.assertRaises(TimeoutError, t.wait_for_messages, {'meteor': ['ready', 'ok', 'ready']}, self.timeout)

        t.wait_for_messages({'rsyslog-forwarder': 'rsyslog'}, self.timeout)
        self.assertRaises(TimeoutError, t.wait_for_messages, {'rsyslog-forwarder': 'ready'}, self.timeout)

        self.assertRaises(UnsupportedError, t.wait_for_messages, {'old': 'ready'}, self.timeout)


class TestStatusMessageMatcher(unittest.TestCase):
    def test_check(self):
        m = StatusMessageMatcher()
        m.check_messages = Mock()
        m.check_set = Mock()
        m.check_list = Mock()

        m.check('ready', ['ready'])
        m.check(re.compile('ready'), ['ready'])
        self.assertEqual(m.check_messages.call_count, 2)

        m.check({'ready'}, ['ready'])
        self.assertEqual(m.check_set.call_count, 1)

        m.check(['ready'], ['ready'])
        m.check(('ready',), ['ready'])
        self.assertEqual(m.check_list.call_count, 2)

    def test_check_messages(self):
        m = StatusMessageMatcher()

        assert m.check_messages('ready', ['ready'])
        assert m.check_messages('ready', ['ready', 'ready'])
        assert not m.check_messages('ready', ['ready', 'ok'])
        assert not m.check_messages('ready', [])

    def test_check_set(self):
        m = StatusMessageMatcher()

        assert m.check_set({'ready'}, ['ready'])
        assert m.check_set({'ready'}, ['ready', 'ok'])
        assert m.check_set({'ready', 'ok'}, ['ready', 'ok'])
        assert not m.check_set({'ready', 'ok'}, ['ready', 'ready'])
        assert not m.check_set({'ok'}, ['ready', 'ready'])
        assert not m.check_set({'ready'}, [])

    def test_check_list(self):
        m = StatusMessageMatcher()
        r = re.compile

        assert m.check_list(['ready'], ['ready'])
        assert not m.check_list(['ready', 'ready'], ['ready'])  # too few
        assert not m.check_list(['ready'], ['ready', 'ready'])  # too many
        assert m.check_list(['ready', 'ready', 'ok'], ['ready', 'ok', 'ready'])
        assert not m.check_list(['ready', 'ready', 'ok'], ['ready', 'ok', 'ok'])
        assert m.check_list([r('ready(ish)?')], ['ready'])
        assert m.check_list([r('ready(ish)?')], ['readyish'])
        assert m.check_list([r('ready(ish)?'), 'ready'], ['ready', 'readyish'])  # ambiguous-ish
        assert m.check_list([r('ready(ish)?'), 'ready'], ['readyish', 'ready'])  # ambiguous-ish

    def test_check_message(self):
        m = StatusMessageMatcher()
        r = re.compile
        self.assertEqual(0, m.check_message('bar', 'foo'))
        self.assertEqual(3, m.check_message('foo', 'foo'))
        self.assertEqual(3, m.check_message(r('foo'), 'foo'))
        self.assertEqual(3, m.check_message(r('foo'), 'foobar'))
        self.assertEqual(3, m.check_message(r('f..'), 'foo'))
        self.assertEqual(0, m.check_message(r('b..'), 'foo'))



import os

from setuptools import setup

version_file = os.path.abspath(
    os.path.join(os.path.dirname(__file__), 'VERSION'))

with open(version_file) as v:
    VERSION = v.read().strip()

install_requires = [
    'requests',
    'libcharmstore',
    'PyYAML',
    'path.py'
]

tests_require = [
    'coverage',
    'nose',
    'pep8',
]


setup(
    name='amulet',
    version=VERSION,
    description='Tools to help with writing Juju Charm Functional tests',
    install_requires=install_requires,
    package_data={'amulet': ['unit-scripts/amulet/*']},
    author='Marco Ceppi',
    author_email='marco@ceppi.net',
    url="https://github.com/juju/amulet",
    packages=['amulet'],
)



import glob
import json
import logging
import os
import subprocess
from datetime import datetime

import pkg_resources

from . import waiter
from . import helpers


# number of seconds an agent must be idle to be considered quiescent
IDLE_THRESHOLD = 30

log = logging.getLogger(__name__)


class SentryError(Exception):
    pass


class Sentry(object):
    def __init__(self, address, port=9001):
        self.config = {}
        self.config['address'] = 'https://%s:%s' % (address, port)

    def file(self, filename):
        return self.file_stat(filename)

    def file_stat(self, filename):
        raise NotImplemented()

    def file_contents(self, filename):
        raise NotImplemented()

    def directory(self, *args):
        return self.directory_stat(*args)

    def directory_stat(self, *args):
        raise NotImplemented()

    def directory_contents(self, *args):
        return self.directory_listing(*args)

    def directory_listing(self, *args):
        raise NotImplemented()

    def juju_agent(self, timeout):
        raise NotImplemented()


class UnitSentry(Sentry):
    """A proxy to a deployed unit, through which the unit can be
    manipulated.

    Provides methods for running commands on the unit and fetching
    relation data from units to which this unit is related.

    :ivar dict info: A dictionary containing 'unit_name' (in the form
        'wordpress/0'), 'service' (name), 'unit' (unit number as string),
        'machine' (machine number as string), 'public-address', and
        'agent-version'.

    """
    @classmethod
    def fromunit(cls, unit):
        pass

    @classmethod
    def fromunitdata(cls, unit, unit_data):
        address = unit_data['public-address']
        unitsentry = cls(address)
        d = unitsentry.info = unit_data
        d['unit_name'] = unit
        d['service'], d['unit'] = unit.split('/')
        unitsentry.upload_scripts()
        return unitsentry

    def upload_scripts(self):
        source = pkg_resources.resource_filename(
            'amulet', os.path.join('unit-scripts', 'amulet'))
        dest = '/tmp/amulet'
        output, code = self.ssh('mkdir -p -m a=rwx {}'.format(dest), raise_on_failure=True)
        # copy one at a time b/c `juju scp -r` doesn't work (currently)
        for f in glob.glob(os.path.join(source, '*.py')):
            cmd = "juju scp {} {}:{}".format(
                os.path.join(source, f),
                self.info['unit_name'], dest)
            subprocess.check_call(cmd.split())

    def _fs_data(self, path):
        return self._run_unit_script("filesystem_data.py {}".format(path))

    def file_stat(self, filename):
        """Run :func:`os.stat` against ``filename`` on the unit.

        :param str filename: Path of file to stat on the remote unit.
        :return: Dictionary containing ``mtime``, ``size``, ``uid``,
            ``gid``, and ``mode`` of ``filename``.

        """
        return self._fs_data(filename)

    def file_contents(self, filename):
        """Get the contents of ``filename`` on the remote unit.

        :param str filename: Path of file to stat on the remote unit.
        :raises: IOError if the call fails.
        :return: File contents as string.

        """
        output, return_code = self._run('cat {}'.format(filename))
        if return_code == 0:
            return output
        else:
            raise IOError(output)

    def directory_stat(self, path):
        """Run :func:`os.stat` against ``path`` on the unit.

        :param str path: Path of directory to stat on the remote unit.
        :return: Dictionary containing ``mtime``, ``size``, ``uid``,
            ``gid``, and ``mode`` of ``path``.

        """
        return self._fs_data(path)

    def directory_listing(self, path):
        """Get the contents of the directory at ``path`` on the remote unit.

        :param str path: Path of directory on the remote unit.
        :return: Dictionary containing 'files' and 'directories', both lists.

        This method does the equivalent of the following, on the remote unit::

            contents = {'files': [], 'directories': []}
            for fd in os.listdir(path):
                if os.path.isfile('{}/{}'.format(path, fd)):
                    contents['files'].append(fd)
                else:
                    contents['directories'].append(fd)
            return contents

        """
        return self._run_unit_script("directory_listing.py {}".format(path))

    def run(self, command):
        """Run an arbitrary command (as root) on the remote unit.

        Uses ``juju run`` to execute the command, which means the command
        will be queued to run after already-queued hooks. To avoid this
        behavior and instead execute the command immediately, see the
        :meth:`ssh` method.

        :param str command: The command to run.
        :return: A 2-tuple containing the output of the command and the exit
            code of the command.

        A default timeout of 5 minutes is imposed on the command. To change
        this timeout, see the :meth:`_run` method.

        """
        output, code = self._run(command)
        return output.strip(), code

    def _run(self, command, unit=None, timeout=300):
        """Run an arbitrary command (as root) on the remote unit.

        Uses ``juju run`` to execute the command, which means the command
        will be queued to run after already-queued hooks. To avoid this
        behavior and instead execute the command immediately, see the
        :meth:`ssh` method.

        :param str command: The command to run.
        :param str unit: Unit on which to run the command, in the form
            'wordpress/0'. If None, defaults to the unit for this
            :class:`UnitSentry`.
        :param int timeout: Seconds to wait before timing out.
        :return: A 2-tuple containing the output of the command and the exit
            code of the command.

        """
        unit = unit or self.info['unit_name']
        cmd = [
            'juju', 'run',
            '--unit', unit,
            '--timeout', "%ds" % timeout,
            command
        ]
        p = subprocess.Popen(
            cmd,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
        )
        stdout, stderr = p.communicate()
        output = stdout if p.returncode == 0 else stderr
        return output.decode('utf8'), p.returncode

    def ssh(self, command, unit=None, raise_on_failure=False):
        """Run an arbitrary command (as the ubuntu user) against a remote
        unit, using `juju ssh`.

        Using `juju ssh` bypasses the Juju execution queue, so it will not
        be blocked by running hooks.  Note, however, that the command is run
        as the ubuntu user instead of root.

        :param str command: The command to run.
        :param str unit: Unit on which to run the command, in the form
            'wordpress/0'. If None, defaults to the unit for this
            :class:`UnitSentry`.
        :param bool raise_on_failure: If True, raises
            :class:`subprocess.CalledProcessError` if the command fails.
        :return: A 2-tuple containing the output of the command and the exit
            code of the command.

        """
        unit = unit or self.info['unit_name']
        cmd = ['juju', 'ssh', unit, command]
        p = subprocess.Popen(
            cmd,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
        )
        stdout, stderr = p.communicate()
        output = stdout if p.returncode == 0 else stderr
        if raise_on_failure and p.returncode != 0:
            raise subprocess.CalledProcessError(p.returncode, cmd, output)
        return output.decode('utf8').strip(), p.returncode

    def _run_unit_script(self, cmd, working_dir=None):
        if working_dir is None:
            working_dir = '/var/lib/juju/agents/unit-{service}-{unit}/charm'.format(**self.info)
        cmd = "/tmp/amulet/{}".format(cmd)
        # XXX: Yes, we are throwing away stderr. Why? Because sudo can write
        # to stderr even when the cmd succeeds. Then, `juju ssh` combines
        # stdout and stderr before handing us the output, so we can't
        # distinguish the two. If we have stderr output mixed in with the
        # stdout of a successful cmd, json parsing will fail.
        output, return_code = self.ssh('cd {} ; sudo {} 2>/dev/null'.format(working_dir, cmd))
        if return_code == 0:
            return json.loads(output)
        else:
            raise IOError(output)

    def juju_agent(self):
        return self._run_unit_script("juju_agent.py", working_dir=".")

    def relation(self, from_rel, to_rel):
        """Get relation data from the remote unit to which we are related,
        denoted by ``to_rel``.

        :param str from_rel: The local side of the relation, e.g. 'website'.
        :param str to_rel: The remote side of the relation,
            e.g. 'haproxy:reverseproxy'.
        :return: Dictionary containing the results of `relation-get`, run
            on the unit on the remote side of the relation.

        """
        this_unit = '{service}/{unit}'.format(**self.info)
        to_service, to_relation = to_rel.split(':')
        r_ids, _ = self._run('relation-ids {}'.format(from_rel))
        r_units = []
        for r_id in r_ids.split():
            r_units.extend(self._run(
                'relation-list -r {}'.format(r_id))[0].split())
        r_units = [u for u in r_units if u.split('/')[0] == to_service]
        for r_unit in r_units:
            r_ids, _ = self._run(
                'relation-ids {}'.format(to_relation), unit=r_unit)
            l_units = []
            for r_id in r_ids.split():
                l_units.extend(self._run(
                    'relation-list -r {}'.format(r_id),
                    unit=r_unit)[0].split())
                if this_unit in l_units:
                    break
            output, _ = self._run(
                'relation-get -r {} - {} --format json'.format(
                    r_id, this_unit), unit=r_unit)
            return json.loads(output)

        raise Exception('Relationship not found')


class Talisman(object):
    """A dict-like object containing a collection of :class:`UnitSentry`
    objects, one for each unit in the deployment.

    Also provides assorted 'wait\_' methods which will block until the
    deployment reaches a certain state.

    See :meth:`__getitem__` for details on retrieving the :class:`UnitSentry`
    objects.

    :note: Under ordinary circumstances this class should not be instantiated
        manually. It should instead be access through the
        :attr:`~amulet.deployer.Deployment.sentry` attribute on an
        :class:`amulet.deployer.Deployment` instance.

    """

    def __init__(self, services, rel_sentry='relation-sentry',
                 juju_env=None, timeout=300):
        self.service_names = services
        self.unit = {}
        self.service = {}

        self.juju_env = juju_env or helpers.default_environment()

        # Save the juju status so we can inspect it later if we don't
        # end up with what we expect in our dictionary of sentries.
        self.status = self.wait_for_status(self.juju_env, services, timeout)

        for service in services:
            if service not in self.status['services']:
                continue  # Raise something?

            service_status = self.status['services'][service]

            if 'units' not in service_status:
                continue

            for unit in service_status['units']:
                unit_data = service_status['units'][unit]
                self.unit[unit] = UnitSentry.fromunitdata(unit, unit_data)
                if 'subordinates' in unit_data:
                    for sub in unit_data['subordinates']:
                        if sub.split('/')[0] not in services:
                            continue
                        subdata = unit_data['subordinates'][sub]
                        self.unit[sub] = UnitSentry.fromunitdata(sub, subdata)

    def __getitem__(self, service):
        """Return the UnitSentry object(s) for ``service``

        :param service: A string in one of two forms::
            "service_name"
            "service_name/unit_num"

        If the first form is used, a list (possibly empty) of all the
        UnitSentry objects for that service is returned.

        If the second form is used, a single object, the UnitSentry for
        the specified unit, is returned.

        Raises a KeyError if the unit does not exist.

        Returns an empty list if the service does not exist or has
        no units.

        Examples::

            >>> d
            <amulet.deployer.Deployment object at 0x7fac83ce8438>
            >>> d.schema()['local']['services']['meteor']['num_units']
            2
            >>> # get UnitSentry for specific unit
            >>> meteor_0 = d.sentry['meteor/0']
            >>> print(meteor_0.info['service'], meteor_0.info['unit'])
            meteor 0
            >>> # get all UnitSentry objects for a service
            >>> for sentry in d.sentry['meteor']:
            ...     print(sentry.info['service'], sentry.info['unit'])
            ...
            meteor 1
            meteor 0
            >>>

        """
        if '/' in service:
            unit_name = service
            return self.unit[unit_name]
        else:
            return [unit_sentry
                    for unit_name, unit_sentry in sorted(self.unit.items())
                    if service == unit_name.split('/', 1)[0]]

    def get_status(self, juju_env=None):
        status = waiter.status(juju_env or self.juju_env)
        machine_states = {}
        normalized = {}
        for number, machine in status['machines'].items():
            machine_states[number] = machine.get('agent-state')
            for container_name, container in machine.get('containers', {}).items():
                machine_states[container_name] = container.get('agent-state')
        for service_name, service in status['services'].items():
            if 'units' not in service and 'relations' not in service:
                # ignore unrelated subordinates; they will never become ready
                continue
            normalized.setdefault(service_name, {})
            for unit_name, unit in service.get('units', {}).items():
                normalized[service_name][unit_name] = {
                    'machine-state': machine_states.get(unit.get('machine')),
                    'public-address': unit.get('public-address'),
                    'workload-status': unit.get('workload-status', {}),
                    'agent-status': unit.get('agent-status', {}),
                    'agent-state': unit.get('agent-state'),
                    'agent-state-info': unit.get('agent-state-info'),
                }
                for sub_name, sub in unit.get('subordinates', {}).items():
                    sub_service = sub_name.split('/')[0]
                    normalized.setdefault(sub_service, {})
                    normalized[sub_service][sub_name] = {
                        'machine-state': machine_states.get(unit.get('machine')),
                        'public-address': sub.get('public-address'),
                        'workload-status': sub.get('workload-status', {}),
                        'agent-status': sub.get('agent-status', {}),
                        'agent-state': sub.get('agent-state'),
                        'agent-state-info': sub.get('agent-state-info'),
                    }
        return normalized

    def wait_for_status(self, juju_env, services, timeout=300):
        """Return environment status, but only after all units have a
        public-address assigned and are in a 'started' state.

        This method is called automatically by :meth:`__init__`, meaning that
        initialization of this :class:`Talisman` object will not complete
        until this method returns. Under ordinary circumstances you should
        not need to call this method manually.

        Raises if a unit reaches error state, or if public-address not
        available for all units before timeout expires.

        :param str juju_env: Name of the juju environment.
        :param dict services: Dictionary of services in the environment.
        :param int timeout: Time to wait before timing out. If environment
            variable AMULET_WAIT_TIMEOUT is set, it overrides this value.
        :return: Dictionary of juju enviroment status.

        """
        timeout = int(os.environ.get('AMULET_WAIT_TIMEOUT') or timeout)

        def check_status(juju_env, services):
            status = self.get_status(juju_env)
            for service_name in services:
                if service_name not in status:
                    # ignore unrelated subordinates; they will never become ready
                    continue
                if not status[service_name]:
                    return False  # expected subordinate
                for unit_name, unit in status[service_name].items():
                    state = unit['workload-status'].get('current') or unit['agent-state']
                    message = unit['workload-status'].get('message') or unit['agent-state-info']
                    if state == 'error':
                        raise Exception('Error on unit {}: {}'.format(
                            unit_name, message))
                    if unit['machine-state'] != 'started':
                        return False
                    if not unit['public-address']:
                        return False
                    # Some substrates (like Amazon) will return a
                    # public-address while the machine is still allocating, so
                    # it's necessary to also check the agent-state to see if
                    # the unit is ready.
                    if unit['agent-state'] not in (None, 'started'):
                        return False
            return True

        for i in helpers.timeout_gen(timeout):
            if check_status(juju_env, services):
                return waiter.status(juju_env)

    def wait(self, timeout=300):
        """Wait for all units to finish running hooks.

        :param int timeout: Number of seconds to wait before timing-out.
            If environment variable AMULET_WAIT_TIMEOUT is set, it overrides
            this value.
        :raises: :class:`amulet.TimeoutError` if the timeout is exceeded.

        """
        timeout = int(os.environ.get('AMULET_WAIT_TIMEOUT') or timeout)

        def check_status():
            status = self.get_status()
            for service_name in self.service_names:
                service = status.get(service_name, {})
                for unit_name, unit in service.items():
                    if unit['agent-status']:
                        if unit['agent-status']['current'] != 'idle':
                            return False
                        since = datetime.strptime(unit['agent-status']['since'][:20], '%d %b %Y %H:%M:%S')
                        if (datetime.now() - since).total_seconds() < IDLE_THRESHOLD:
                            return False
                    else:
                        running_hooks = self.unit[unit_name].juju_agent()
                        if running_hooks is None or running_hooks:
                            return False
            return True

        log.info('Waiting up to %s seconds for deployment to settle...',
                 timeout)
        start = datetime.now()
        for i in helpers.timeout_gen(timeout):
            if check_status():
                log.info('Deployment settled in %s seconds.',
                         (datetime.now() - start).total_seconds())
                return

    def wait_for_messages(self, messages, timeout=300):
        """Wait for specific extended status messages to be set via status-set.

        Note that if this is called on an environment that doesn't support
        extended status (pre Juju 1.24), it will raise an
        :class:`~amulet.helpers.UnsupportedError` exception.

        :param dict messages: A mapping of services to an exact message,
            a regular expression, a set of messages or regular expressions,
            or a list of messages or regular expressions.  If a single message
            is given, all units of the service must match.  If a set is given,
            then each message in the set must match at least one unit.  If a
            list is given, then there must be a one-to-one match between the
            messages and the units, though the order doesn't matter.
        :param int timeout: Number of seconds to wait before timing-out.

        Examples::

            # wait for all units to report "ready"
            t.wait_for_messages({'ubuntu': 'ready'})

            # wait for all units to report something like "ready"
            t.wait_for_messages({'ubuntu': re.compile('r..dy')})

            # wait for at least one unit to report "ready"
            t.wait_for_messages({'ubuntu': {'ready'}})

            # wait for all units to report either "ready" or "ok"
            t.wait_for_messages({'ubuntu': re.compile('ready|ok')})

            # wait for at least one unit to report "ready" and at least one
            # unit to report "ok"
            t.wait_for_messages({'ubuntu': {'ready', 'ok'}})

            # wait for one unit to report "ready" and the other to report "ok"
            # (must be exactly two units)
            t.wait_for_messages({'ubuntu': ['ready', 'ok']})

        """
        def get_messages(service, status):
            messages = []
            for unit in status.get(service, {}).values():
                if not unit['workload-status']:
                    raise helpers.UnsupportedError()
                messages.append(unit['workload-status'].get('message', ''))
            return messages

        matcher = StatusMessageMatcher()
        for i in helpers.timeout_gen(timeout):
            status = self.get_status()
            for service, expected in messages.items():
                actual = get_messages(service, status)
                if not matcher.check(expected, actual):
                    break
            else:
                return

    def _sync(self):
        pass


class StatusMessageMatcher(object):
    def check(self, expected, actual):
        if isinstance(expected, (list, tuple)):
            return self.check_list(expected, actual)
        elif isinstance(expected, set):
            return self.check_set(expected, actual)
        else:
            return self.check_messages(expected, actual)

    def check_messages(self, expected, actual):
        """
        Check a single string or regexp against a list of messages.

        All messages must match the string or regexp.
        """
        if not actual:
            return False
        for a in actual:
            if not self.check_message(expected, a):
                return False
        return True

    def check_set(self, expected, actual):
        """
        Check a set of strings or regexps against a list of messages.

        Each expected string or regexp must match at least once.
        """
        if not actual:
            return False
        for e in expected:
            for a in actual:
                if self.check_message(e, a):
                    break  # match, go to next expected
            else:
                return False  # no match for this expected (all must match)
        return True

    def check_list(self, expected, actual):
        """
        Check a list of strings or regexps against a list of messages.

        Each expected string or regexp must match once, and all messages must be matched.
        If an expected string or regexp matches multiple messages, longer matches are
        preferred, to resolve the ambiguity.
        """
        if len(actual) != len(expected):
            return False
        actual = list(actual)  # copy
        for e in expected:
            im = None
            m = 0
            for i, a in enumerate(actual):
                n = self.check_message(e, a)
                if n > m:  # prefer longest matches
                    im = i
                    m = n
            if im is None:
                return False  # no matches for this expected (all must match)
            actual.pop(im)  # remove matched to ensure 1-to-1
        return True

    def check_message(self, expected, actual):
        """
        Check a single string or regexp against a single message.

        Returns the length of the match (0 for no match), to allow for preferring longer matches.
        """
        if hasattr(expected, 'search'):
            m = expected.search(actual)
            return len(m.group()) if m else 0
        elif expected == actual:
            return len(actual)
        else:
            return 0


class ServiceSentry(Sentry):
    pass
